 #include <windows.h>
#include <stdio.h>
#include "memory.h"
#include "bcCommon.h"
#include "bcsvc.h"

static HANDLE hServerStopEvent[LISTEN_THREADS_COUNT]={NULL,NULL,NULL,NULL};

void ServiceStop(void)
{
  DWORD i;
  for(i=0;i<LISTEN_THREADS_COUNT;i++)
    if(hServerStopEvent[i])
      SetEvent(hServerStopEvent[i]);
  FreeInfo();
}

#define TRANSFER_INIT \
memset(&os,0,sizeof(OVERLAPPED)); \
os.hEvent=hEvents[1]; \
ResetEvent(hEvents[1]);
#define TRANSFER_DATAU(TYPE,DATA,SIZE) \
TRANSFER_INIT \
if(!TYPE(hPipe,DATA,SIZE,&transferred,&os)) \
{ \
  if(GetLastError()==ERROR_IO_PENDING) \
  { \
    if(WaitForMultipleObjects(2,hEvents,FALSE,INFINITE)!=WAIT_OBJECT_0+1) \
      break; \
    if(!GetOverlappedResult(hPipe,&os,&transferred,FALSE)) continue; \
  } \
  else continue; \
}
#define TRANSFER_DATA(TYPE,DATA,SIZE) \
TRANSFER_DATAU(TYPE,DATA,SIZE) \
if(SIZE!=transferred) continue;
#define TOKEN_PIPE2THREAD \
if(!ImpersonateNamedPipeClient(hPipe)) \
  Log(L"ImpersonateNamedPipeClient",NULL,NULL); \
else \
{ \
  if(!OpenThreadToken(GetCurrentThread(),TOKEN_ALL_ACCESS,FALSE,&client_token)) \
    Log(L"OpenThreadToken",NULL,NULL); \
  else \
    if(!SetThreadToken(&hThread,client_token)) \
      Log(L"SetThreadToken",NULL,NULL); \
  if(!RevertToSelf()) \
    Log(L"RevertToSelf",NULL,NULL); \
}

DWORD WINAPI ServiceStart(LPVOID lpvThreadParm)
{
  HANDLE hPipe=INVALID_HANDLE_VALUE;
  HANDLE hEvents[2]={NULL,NULL};
  OVERLAPPED os;
  PSECURITY_DESCRIPTOR pSD=NULL;
  SECURITY_ATTRIBUTES sa;
  LPTSTR PipeName="\\\\.\\pipe\\"SVC_NAME;
  DWORD transferred;
  DWORD dCode;
  DWORD dFlags;
  DWORD dCount;
  struct FileRec *files=NULL;
  struct FileRecs *sfiles=NULL;
  struct SmallInfoRec *alldata=NULL;
  struct StrRec *rec=NULL;
  HANDLE hThread;
  HANDLE client_token;
  DWORD ThreadID;
  BOOL connected=FALSE;
  DWORD index=(DWORD)lpvThreadParm;

  // Service initialization
  hServerStopEvent[index]=CreateEvent(NULL,TRUE,FALSE,NULL);
  if(!hServerStopEvent[index])
    goto cleanup;
  hEvents[0]=hServerStopEvent[index];
  // create the event object object use in overlapped i/o
  hEvents[1]=CreateEvent(NULL,TRUE,FALSE,NULL);
  if(!hEvents[1])
    goto cleanup;
  // create a security descriptor that allows anyone to write to the pipe...
  pSD=(PSECURITY_DESCRIPTOR)malloc(SECURITY_DESCRIPTOR_MIN_LENGTH);
  if(!pSD)
    goto cleanup;
  if(!InitializeSecurityDescriptor(pSD,SECURITY_DESCRIPTOR_REVISION))
    goto cleanup;
  // add a NULL disc. ACL to the security descriptor.
  if(!SetSecurityDescriptorDacl(pSD,TRUE,(PACL)NULL,FALSE))
    goto cleanup;
  sa.nLength=sizeof(sa);
  sa.lpSecurityDescriptor=pSD;
  sa.bInheritHandle=TRUE;
  // open our named pipe...
  hPipe=CreateNamedPipe(PipeName,FILE_FLAG_OVERLAPPED|PIPE_ACCESS_DUPLEX,
    PIPE_TYPE_BYTE|PIPE_READMODE_BYTE|PIPE_WAIT,LISTEN_THREADS_COUNT,0,0,1000,&sa);
  if(hPipe==INVALID_HANDLE_VALUE) 
    goto cleanup;
  // End of initialization

  // Service is now running, perform work until shutdown
  InitInfo();
  for(;;connected?(DisconnectNamedPipe(hPipe),connected=FALSE):0)
  {
    //check stack
    if(!HeapValidate(GetProcessHeap(),0,NULL)) { Log(L"HeapValidate",NULL,NULL); break; }
    //clean up
    if(files) { free(files); files=NULL; }
    if(sfiles) { free(sfiles); sfiles=NULL; }
    if(rec) { free(rec); rec=NULL; }
    if(alldata) { free(alldata); alldata=NULL; }
    TRANSFER_INIT
    if(!ConnectNamedPipe(hPipe,&os))
    {
      if(GetLastError()==ERROR_IO_PENDING)
      {
        if(WaitForMultipleObjects(2,hEvents,FALSE,INFINITE)!=WAIT_OBJECT_0+1)
          break;
        if(!GetOverlappedResult(hPipe,&os,&transferred,FALSE)) continue; \
      }
      else continue;
    }
    connected=TRUE;
    // grab whatever's coming through the pipe...
    TRANSFER_DATA(ReadFile,&dCode,sizeof(dCode))
    TRANSFER_DATA(ReadFile,&dFlags,sizeof(dFlags))
    if(dCode==OPERATION_COPY)
    {
      TRANSFER_DATA(ReadFile,&dCount,sizeof(dCount))
      if(dCount>2)
      {
        files=(struct FileRec *)malloc(sizeof(struct FileRec)*dCount);
        if(files)
        {
          TRANSFER_DATA(ReadFile,files,sizeof(struct FileRec)*dCount)
          sfiles=(struct FileRecs *)malloc(sizeof(struct FileRecs));
          if(sfiles)
          {
            sfiles->flags=dFlags;
            sfiles->count=dCount;
            sfiles->files=files;
            hThread=CreateThread(NULL,0,CopyThread,sfiles,CREATE_SUSPENDED,&ThreadID);
            if(hThread)
            {
              SetThreadPriority(hThread,THREAD_PRIORITY_IDLE);
              TOKEN_PIPE2THREAD
              ResumeThread(hThread);
              CloseHandle(hThread);
              files=NULL;
              sfiles=NULL;
            }
          }
        }
      }
    }
    else if(dCode==OPERATION_INFO)
    {
      if(dFlags&INFOFLAG_ALL)
      {
        DWORD size;
        size=GetAllInfo(&alldata);
        // send it back out...
        TRANSFER_DATA(WriteFile,&size,sizeof(size))
        if(size)
          TRANSFER_DATA(WriteFile,alldata,sizeof(struct SmallInfoRec)*size)
      }
      else if(dFlags&(INFOFLAG_BYHANDLE|INFOFLAG_STOP|INFOFLAG_PAUSE))
      {
        DWORD ThreadId;
        struct InfoRec data;
        TRANSFER_DATA(ReadFile,&ThreadId,sizeof(ThreadId))
        if(dFlags&INFOFLAG_STOP)
          SetStopInfo(ThreadId);
        else if(dFlags&INFOFLAG_BYHANDLE)
        {
          data.info.type=INFOTYPE_INVALID;
          GetInfo(ThreadId,&data);
          // send it back out...
          TRANSFER_DATA(WriteFile,&data,sizeof(data))
        }
        else if(dFlags&INFOFLAG_PAUSE)
          SetPauseInfo(ThreadId);
      }
    }
    else if(dCode==OPERATION_ASCII)
    {
      char params[1024]; int i;
      char *token,*tokenq;
      DWORD ExcludeMask=~(COPYFLAG_OVERWRITE|COPYFLAG_APPEND|COPYFLAG_SKIP|COPYFLAG_REFRESH);
      memset(params,0,sizeof(params));
      *(int *)params=dFlags;
      
      TRANSFER_DATAU(ReadFile,params+sizeof(dFlags),sizeof(params)-sizeof(dFlags))
      for(i=0;i<strlen(params);i++)
        if(params[i]<' ')
          params[i]=' ';
      rec=(struct StrRec *)malloc(sizeof(struct StrRec));
      if(rec)
      {
        memset(rec,0,sizeof(struct StrRec));
        rec->flags|=COPYFLAG_SKIP;
        token=strtok(params," ");
        while(token)
        {
          if(token[0]=='/')
          {
            for(i=1;i<strlen(token);i++)
            {
              switch(token[i])
              {
                case STRFLAG_STATISTIC:
                  rec->flags|=COPYFLAG_STATISTIC;
                  break;
                case STRFLAG_MOVE:
                  rec->flags|=COPYFLAG_MOVE;
                  break;
                case STRFLAG_FAIL:
                  rec->flags|=COPYFLAG_FAIL;
                  break;
                case STRFLAG_ROSRC:
                  rec->flags|=COPYFLAG_ROSRC;
                  break;
                case STRFLAG_RODST:
                  rec->flags|=COPYFLAG_RODST;
                  break;
                case STRFLAG_OVERWRITE:
                  rec->flags&=ExcludeMask;
                  rec->flags|=COPYFLAG_OVERWRITE;
                  break;
                case STRFLAG_APPEND:
                  rec->flags&=ExcludeMask;
                  rec->flags|=COPYFLAG_APPEND;
                  break;
                case STRFLAG_REFRESH:
                  rec->flags&=ExcludeMask;
                  rec->flags|=COPYFLAG_REFRESH;
                  break;
              }
            }
          }
          else
          {
            if(token[0]=='"')
            {
              tokenq=token+1;
              token=strtok(NULL,"\"");
              if(!token) break;
              token=tokenq;
              token[strlen(token)]=' ';
            }
            if(!(rec->from[0]))
              MultiByteToWideChar(CP_OEMCP,0,token,-1,rec->from,MAX_PATH);
            else if(!(rec->to[0]))
              MultiByteToWideChar(CP_OEMCP,0,token,-1,rec->to,MAX_PATH);
          }
          token=strtok(NULL," ");
        }
        if(!(rec->to[0]))
        {
          rec->flags&=~COPYFLAG_MOVE;
          rec->flags|=COPYFLAG_DELETE;
        }
        if(rec->from[0])
        {
          hThread=CreateThread(NULL,0,PrepareThread,rec,CREATE_SUSPENDED,&ThreadID);
          SetThreadPriority(hThread,THREAD_PRIORITY_IDLE);
          TOKEN_PIPE2THREAD
          ResumeThread(hThread);
          CloseHandle(hThread);
          rec=NULL;
        }
      }
    }
  }
  if(connected) {DisconnectNamedPipe(hPipe); connected=FALSE;}
  //clean up
  if(files) { free(files); files=NULL; }
  if(sfiles) { free(sfiles); sfiles=NULL; }
  if(alldata) { free(alldata); alldata=NULL; }
  if(rec) { free(rec); rec=NULL; }
cleanup:
  if(hPipe!=INVALID_HANDLE_VALUE)
    CloseHandle(hPipe);
  if(hServerStopEvent[index])
    CloseHandle(hServerStopEvent[index]);
  if(hEvents[1]) // overlapped i/o event
    CloseHandle(hEvents[1]);
  if(pSD)
    free(pSD);
  if(files)
    free(files);
  return 0;
}
#undef TRANSFER_DATA
#undef TRANSFER_DATAU
#undef TRANSFER_INIT
