#include <windows.h>
#include <stdio.h>
#include "bcversion.h"
#include "bcCommon.h"
#include "bcsvc.h"

static SERVICE_STATUS_HANDLE ServiceStatusHandle;
static SERVICE_STATUS ServiceStatus;
static HANDLE threads[LISTEN_THREADS_COUNT];

BOOL AddEventSupport(void)
{
  HKEY hk; 
  DWORD dwData; 
  char filename[MAX_PATH];
  if(!GetModuleFileName(NULL,filename,MAX_PATH))
    return FALSE;
  /* 
   * Add your source name as a subkey under the Application 
   * key in the EventLog service portion of the registry. 
   */ 
  if(RegCreateKey(HKEY_LOCAL_MACHINE,"SYSTEM\\CurrentControlSet\\Services\\EventLog\\Application\\"SVC_DISPLAYNAMEA,&hk)) 
    return FALSE; 
  /* Add the Event ID message-file name to the subkey. */ 
  if(RegSetValueEx(hk,"EventMessageFile",0,REG_EXPAND_SZ,(LPBYTE)filename,strlen(filename)+1))
    return FALSE; 
  /* Set the supported types flags. */ 
  dwData=EVENTLOG_ERROR_TYPE|EVENTLOG_WARNING_TYPE|EVENTLOG_INFORMATION_TYPE; 
  if(RegSetValueEx(hk,"TypesSupported",0,REG_DWORD,(LPBYTE) &dwData,sizeof(DWORD)))
    return FALSE; 
  RegCloseKey(hk);
  return TRUE;
}

BOOL ReportStatusToSCMgr(DWORD State,DWORD ExitCode,DWORD Wait)
{
  BOOL ret=TRUE;
  if(State==SERVICE_START_PENDING)
    ServiceStatus.dwControlsAccepted=0;
  else
    ServiceStatus.dwControlsAccepted=SERVICE_ACCEPT_STOP|SERVICE_ACCEPT_SHUTDOWN;
  ServiceStatus.dwCurrentState=State;
  ServiceStatus.dwWin32ExitCode=ExitCode;
  ServiceStatus.dwWaitHint=Wait;
  if ((State==SERVICE_RUNNING) || (State==SERVICE_STOPPED))
    ServiceStatus.dwCheckPoint=0;
  else
    ServiceStatus.dwCheckPoint++;
  if(!(ret=SetServiceStatus(ServiceStatusHandle,&ServiceStatus)))
    Log(L"SetServiceStatus",NULL,NULL);
  return ret;
}

void WINAPI Handler(DWORD fdwControl)
{
  DWORD i;
  switch(fdwControl)
  {
    case SERVICE_CONTROL_SHUTDOWN:
    case SERVICE_CONTROL_STOP:
      ReportStatusToSCMgr(SERVICE_STOP_PENDING,NO_ERROR,30000);
      ServiceStop();
      WaitForMultipleObjects(LISTEN_THREADS_COUNT,threads,TRUE,INFINITE);
      for(i=0;i<LISTEN_THREADS_COUNT;i++)
        CloseHandle(threads[i]);
      ServiceStatus.dwCurrentState=SERVICE_STOPPED;
      break;
  }
  ReportStatusToSCMgr(ServiceStatus.dwCurrentState,NO_ERROR,0);
}

void WINAPI ServiceMain(DWORD dwArgc, LPSTR *lpszArgv)
{
  DWORD ThreadID,i;
  ServiceStatusHandle=RegisterServiceCtrlHandlerA(SVC_NAME, Handler);
  if(!ServiceStatusHandle)
  {
    ReportStatusToSCMgr(SERVICE_STOPPED,GetLastError(),0);
    return;
  }
  ServiceStatus.dwServiceType=SERVICE_WIN32_OWN_PROCESS;
  ServiceStatus.dwServiceSpecificExitCode=0;
  ServiceStatus.dwCheckPoint=1;
  if(!ReportStatusToSCMgr(SERVICE_START_PENDING,NO_ERROR,3000))
  {
    ReportStatusToSCMgr(SERVICE_STOPPED,GetLastError(),0);
    return;
  }
  for(i=0;i<LISTEN_THREADS_COUNT;i++)
    threads[i]=CreateThread(NULL,0,ServiceStart,(void *)i,0,&ThreadID);
  ReportStatusToSCMgr(SERVICE_RUNNING,NO_ERROR,0);
}


void InstallService(int Type)
{
  char path[1024];
  char dependencies[2]={0,0};
  SC_HANDLE ServiceControlHandle;
  SC_HANDLE SCManagerHandle;

  if(!AddEventSupport())
    fprintf(stderr,"Can't create Event Log support value in registry (%ld).\n",GetLastError());
   
  if(!GetModuleFileName(0, path, sizeof(path)))
  {
    fprintf(stderr, "Unable to install "SVC_NAME", GetModuleFileName Fail.\n");
    return;
  }

  SCManagerHandle=OpenSCManager(NULL,NULL,SC_MANAGER_ALL_ACCESS);
  if(SCManagerHandle>0)
  {
    ServiceControlHandle=CreateServiceA(SCManagerHandle,SVC_NAME,
      SVC_DISPLAYNAMEA,SERVICE_ALL_ACCESS,SERVICE_WIN32_OWN_PROCESS,
      Type,SERVICE_ERROR_NORMAL,path,NULL,NULL,dependencies,NULL,NULL);
    if(ServiceControlHandle>0)
    {
      fprintf(stderr, "Install Ok.\n");
      CloseServiceHandle(ServiceControlHandle);
    }
    else
      fprintf(stderr, "Unable to install "SVC_NAME", CreateService Fail.\n");
  }
  else
  {
    fprintf(stderr, "Unable to install "SVC_NAME", OpenSCManager Fail.\n");
  }
}

void UnInstallService(void)
{
  SC_HANDLE ServiceControlHandle;
  SC_HANDLE SCManagerHandle;

  SCManagerHandle=OpenSCManager(NULL,NULL,SC_MANAGER_ALL_ACCESS);
  if(SCManagerHandle>0)
  {
    ServiceControlHandle=OpenService(SCManagerHandle,SVC_NAME,SERVICE_ALL_ACCESS);
    if(ServiceControlHandle>0)
    {
      if(ControlService(ServiceControlHandle,SERVICE_CONTROL_STOP,&ServiceStatus))
      {
        fprintf(stderr, "Stopping Service ");
        Sleep(1000);
        while(QueryServiceStatus(ServiceControlHandle,&ServiceStatus))
        {
          if(ServiceStatus.dwCurrentState==SERVICE_STOP_PENDING)
          {
            fprintf(stderr, ".");
            Sleep(1000);
          }
          else
            break;
        }
        fprintf(stderr, "\n");

        if(ServiceStatus.dwCurrentState==SERVICE_STOPPED)
          fprintf(stderr, "Service Stop Now\n");
        else
        {
          CloseServiceHandle(ServiceControlHandle);
          CloseServiceHandle(SCManagerHandle);
          fprintf(stderr, "Service Stop Fail\n");
          return;
        }
      }
      if(DeleteService(ServiceControlHandle))
        fprintf(stderr, "Service Uninstall Ok.\n");
      else
        fprintf(stderr, "DeleteService fail (%ld).\n",GetLastError());
      CloseServiceHandle(ServiceControlHandle);
    }
    else
      fprintf(stderr, "OpenService fail (%ld).\n",GetLastError());
    CloseServiceHandle(SCManagerHandle);
  }
  else
    fprintf(stderr, "OpenSCManager fail (%ld).\n",GetLastError());
}

int main(int argc,char *argv[])
{
  SERVICE_TABLE_ENTRYA ste[] = {{SVC_NAME,ServiceMain},{NULL,NULL}};
  char progNameReally[1024];
  char *progName, *tmp;

  if(argc>1)
  {
    if(!stricmp(argv[1],"/?"))
    {
      strcpy(progNameReally, argv[0]);
      progName = &progNameReally[0];
      for (tmp = &progNameReally[0]; *tmp != '\0'; tmp++)
        if (*tmp == '\\') progName = tmp + 1;
      fprintf(
        stderr,
        VERSION
        "\n   usage: %s [flags]\n"
        "   \n"
        "   /I install the service for manual startup\n"
        "   /A install the service for automatic startup\n"
        "   /D install the service to be installed as disabled\n"
        "   /R uninstall the service\n"
        "   /? help\n"
        "   \n",
        progName
      );
    }
    else if(!stricmp(argv[1],"/I"))
    {
      InstallService(SERVICE_DEMAND_START);
    }
    else if(!stricmp(argv[1],"/A"))
    {
      InstallService(SERVICE_AUTO_START);
    }
    else if(!stricmp(argv[1],"/D"))
    {
      InstallService(SERVICE_DISABLED);
    }
    else if(!stricmp(argv[1],"/R"))
    {
      UnInstallService();
    }
  }
  else
  {
    StartServiceCtrlDispatcher(ste);
  }
  return 0;
}